""" Add convenience functions here if needed
"""

__author__ = 'noe'

import numpy as _np

class ZeroRankError(_np.linalg.LinAlgError):
    """Input matrix has rank zero."""
    pass

def features_to_basis(infiles, basisset, outfiles):
    """Reads input files

    basisset : BasisSet object
        basis set tob e used

    References
    ---------
    .. [5] Vitalini, F., Noe, F. and Keller, B. (2015):
        A basis set for peptides for the variational approach to conformational kinetics. (In review).

    """
    # cycle through input files
    # read infile
    # map to basis function values
    # write outfile
    pass
