#!C:\bld\pyemma_1550861281410\_h_env\python.exe
# EASY-INSTALL-ENTRY-SCRIPT: 'pyEMMA==2.5.5','console_scripts','pyemma_list_models'
__requires__ = 'pyEMMA==2.5.5'
import re
import sys
from pkg_resources import load_entry_point

if __name__ == '__main__':
    sys.argv[0] = re.sub(r'(-script\.pyw?|\.exe)?$', '', sys.argv[0])
    sys.exit(
        load_entry_point('pyEMMA==2.5.5', 'console_scripts', 'pyemma_list_models')()
    )
