from glob import glob

import tables
import h5py

import os

PRE_CALC_FEATURES_PATH = '/group/ag_cmb/marscher/feature_sel/'

files = glob(PRE_CALC_FEATURES_PATH + '/*.h5')
assert files

to_delete = []
for f in files:
    break
    with h5py.File(f, 'r') as fh:
        if len(fh.keys()) == 0:
            to_delete.append(f)

print(to_delete)

rm = ['/group/ag_cmb/marscher/feature_sel/A3D-0-protein-013.dcd_backbone.h5',
      '/group/ag_cmb/marscher/feature_sel/A3D-1-protein-121.dcd_backbone.h5',
      '/group/ag_cmb/marscher/feature_sel/A3D-1-protein-151.dcd_backbone.h5',
      '/group/ag_cmb/marscher/feature_sel/A3D-1-protein-159.dcd_backbone.h5',
      '/group/ag_cmb/marscher/feature_sel/A3D-0-protein-129.dcd_backbone.h5']

rm = ['/group/ag_cmb/marscher/feature_sel/GTT-0-protein-203.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-079.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-233.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-214.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-066.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-209.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-319.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-143.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-174.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-160.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-031.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-018.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-061.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-191.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-147.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-168.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-052.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-320.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-226.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-325.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-294.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-075.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-308.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-221.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-119.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-229.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-127.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-108.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-015.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-203.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-281.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-059.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-276.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-004.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-029.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-090.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-237.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-040.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-145.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-097.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-118.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-279.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-303.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-114.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-212.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-157.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-190.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-100.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-210.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-035.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-222.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-225.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-221.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-261.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-014.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-201.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-224.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-165.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-158.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-068.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-057.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-101.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-284.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-273.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-191.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-310.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-133.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-251.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-105.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-227.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-207.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-188.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-216.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-234.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-304.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-259.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-169.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-001.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-256.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-048.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-020.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-152.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-264.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-153.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-028.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-302.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-179.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-157.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-197.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-167.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-102.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-212.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-008.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-211.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-300.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-186.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-095.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-175.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-106.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-055.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-116.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-254.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-135.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-287.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-204.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-265.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-033.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-280.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-170.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-188.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-159.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-006.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-163.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-182.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-223.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-271.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-091.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-049.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-241.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-051.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-142.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-318.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-230.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-200.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-196.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-190.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-047.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-231.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-235.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-215.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-136.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-242.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-315.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-138.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-173.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-178.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-292.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-013.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-298.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-130.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-053.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-073.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-081.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-312.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-194.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-208.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-267.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-044.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-240.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-219.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-244.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-208.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-062.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-213.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-184.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-295.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-113.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-286.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-132.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-185.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-148.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-162.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-104.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-323.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-002.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-313.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-000.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-178.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-046.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-111.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-151.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-285.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-206.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-195.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-199.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-093.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-250.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-255.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-180.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-110.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-243.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-234.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-214.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-249.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-163.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-094.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-024.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-213.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-071.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-238.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-165.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-050.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-043.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-317.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-065.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-058.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-023.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-171.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-253.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-232.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-034.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-322.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-176.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-139.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-083.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-270.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-140.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-228.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-210.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-187.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-283.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-222.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-172.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-084.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-215.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-324.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-164.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-177.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-220.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-277.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-299.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-181.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-016.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-189.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-005.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-236.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-205.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-122.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-082.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-241.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-154.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-193.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-168.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-092.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-150.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-248.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-209.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-078.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-089.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-278.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-167.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-022.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-103.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-125.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-038.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-027.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-211.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-039.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-217.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-227.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-232.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-195.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-258.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-010.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-174.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-171.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-060.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-305.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-262.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-137.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-291.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-238.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-141.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-076.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-054.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-126.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-164.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-072.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-129.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-069.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-099.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-199.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-032.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-245.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-282.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-183.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-201.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-107.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-219.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-196.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-252.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-189.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-041.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-217.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-289.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-169.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-128.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-056.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-202.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-070.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-030.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-207.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-045.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-204.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-229.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-067.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-124.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-301.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-236.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-161.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-275.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-170.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-269.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-063.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-007.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-231.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-146.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-309.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-235.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-218.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-155.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-240.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-237.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-263.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-268.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-086.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-186.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-257.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-272.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-064.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-175.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-003.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-297.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-019.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-243.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-077.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-120.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-026.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-172.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-194.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-042.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-112.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-246.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-085.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-228.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-307.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-197.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-180.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-198.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-288.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-266.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-225.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-173.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-296.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-290.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-192.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-021.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-098.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-316.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-087.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-242.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-144.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-224.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-088.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-260.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-011.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-230.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-183.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-200.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-185.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-017.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-121.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-131.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-036.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-149.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-233.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-096.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-012.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-080.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-166.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-176.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-198.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-223.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-274.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-226.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-025.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-321.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-134.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-009.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-306.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-247.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-117.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-109.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-166.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-115.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-293.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-220.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-177.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-205.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-187.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-239.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-314.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-193.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-123.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-206.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-202.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-311.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-216.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-037.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-074.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-239.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-192.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-184.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-218.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-156.dcd_dssp.h5']

for f in rm:
    os.remove(f)
